'**************************************************
' FILE:         Encoding
' AUTHOR:       Paulo Santos
' CREATED:      2005.JAN.13
' COPYRIGHT:    Copyright 2005-2007  Paulo Santos
'               All Rights Reserved.
' DESCRIPTION:
'       Encoding Function
'
' MODIFICATION HISTORY:
' 01    2005.JAN.13
'       Paulo Santos
'       Added a New Encoding Class
' 
' 02    2005.JAN.14
'       Paulo Santos
'       Added summarized description of 
'       each component.
'
' 03    2005.FEB.20
'       Paulo Santos
'       Added an enhanced GetBytes function
'
' 04    2005.FEB.21
'       Paulo Santos
'       Added an enhanced GetChars function
'
' 05    2005.DEC.29
'       Paulo Santos
'       Updated to .NET 2.0
'
' 06    2007.OCT.09
'       Paulo Santos
'       Updated documentation
'       Refactored to enhance performance
'***************************************************

Namespace System.Text

    ''' <summary>
    ''' Helps to convert a ANSI input into a ANSI string.
    ''' </summary>
    Public Class ANSIEncoding
        Inherits System.Text.Encoding

        ''' <summary>
        ''' It is used to control an wait event for the GetBytes function.
        ''' </summary>
        Private __ThreadsGetBytes As Integer
        ''' <summary>
        ''' It is used to control an wait event for the GetChars function.
        ''' </summary>
        Private __ThreadsGetChars As Integer

        ''' <summary>
        ''' The stateObjectStruct contains the data for the work threads that converts the characters.
        ''' </summary>
        Private Structure stateObjectStruct
            ''' <summary>
            ''' The character array.
            ''' </summary>
            ''' <remarks></remarks>
            Dim charArray() As Char
            ''' <summary>
            ''' The byte array.
            ''' </summary>
            ''' <remarks></remarks>
            Dim byteArray() As Byte
            ''' <summary>
            ''' The index of the byte array needed for the GetBytes or GetChars.
            ''' </summary>
            ''' <remarks></remarks>
            Dim byteIndex As Integer
            ''' <summary>
            ''' The index of the char array needed for the GetBytes or GetChars.
            ''' </summary>
            ''' <remarks></remarks>
            Dim charIndex As Integer
            ''' <summary>
            ''' The count of bytes or chars.
            ''' </summary>
            ''' <remarks></remarks>
            Dim Count As Integer
        End Structure

        ''' <summary>
        ''' Initializes a new instance of the ANSI Encoding class.
        ''' </summary>
        <DebuggerStepThrough()> _
        Public Sub New()
            MyBase.New()
        End Sub

        ''' <summary>Gets the human-readable description of the current encoding.</summary>
        ''' <returns>The human-readable description of the current <see cref="T:System.Text.Encoding"></see>.</returns>
        Public Overrides ReadOnly Property EncodingName() As String
            <DebuggerHidden()> _
            Get
                Return "ANSI"
            End Get
        End Property

        ''' <summary>Calculates the number of bytes produced by encoding a set of characters from the specified character array.</summary>
        ''' <returns>The number of bytes produced by encoding the specified characters.</returns>
        ''' <param name="chars">The character array containing the set of characters to encode. </param>
        ''' <param name="count">The number of characters to encode. </param>
        ''' <param name="index">The index of the first character to encode. </param>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">index or count is less than zero.
        ''' <para>- or -</para>index and count do not denote a valid range in chars.</exception>
        ''' <exception cref="T:System.ArgumentNullException">chars is <langword name="null"/>.</exception>
        ''' <remarks>
        ''' The ANSIEncoding class uses a direct conversion each character is ONE byte.
        ''' </remarks>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetByteCount(ByVal chars() As Char, ByVal index As Integer, ByVal count As Integer) As Integer

            '*
            '* Check all the parameters
            '*
            If (chars Is Nothing) Then
                Throw New ArgumentNullException("chars")
            End If

            If (index < 0) Then
                Throw New ArgumentOutOfRangeException("index", index, String.Format(My.Resources.CannotBeNegative, "index"))
            End If

            If (count < 0) Then
                Throw New ArgumentOutOfRangeException("count", count, String.Format(My.Resources.CannotBeNegative, "count"))
            End If

            If (index + count > chars.Length) Then
                Throw New ArgumentOutOfRangeException("", index + count, String.Format(My.Resources.IndexAndCountAreNotValidInArray, "index", "count", "chars"))
            End If

            Return (count)
        End Function

        ''' <summary>Encodes a set of characters from the specified character array into the specified byte array.</summary>
        ''' <returns>The actual number of bytes written into bytes.</returns>
        ''' <param name="chars">The character array containing the set of characters to encode. </param>
        ''' <param name="charIndex">The index of the first character to encode. </param>
        ''' <param name="charCount">The number of characters to encode. </param>
        ''' <param name="byteIndex">The index at which to start writing the resulting sequence of bytes. </param>
        ''' <param name="bytes">The byte array to contain the resulting sequence of bytes. </param>
        ''' <exception cref="T:System.ArgumentNullException">chars is <langword name="null"/>.
        ''' <para>- or -</para>bytes is <langword name="null"/>.</exception>
        ''' <exception cref="T:System.ArgumentException">bytes does not have enough capacity from byteIndex to the end of the array to accommodate the resulting bytes. </exception>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">charIndex or charCount or byteIndex is less than zero.
        ''' <para>- or -</para>charIndex and charCount do not denote a valid range in chars.
        ''' <para>- or -</para>byteIndex is not a valid index in bytes.</exception>
        ''' <remarks>
        ''' Each char in the array is converted to a byte through the <see cref="M:Microsoft.VisualBasic.Asc"/> function.
        ''' </remarks>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetBytes(ByVal chars() As Char, ByVal charIndex As Integer, ByVal charCount As Integer, ByVal bytes() As Byte, ByVal byteIndex As Integer) As Integer

            '*
            '* Check all the parameters
            '*
            If (bytes Is Nothing) Then
                Throw New ArgumentNullException("bytes")
            End If

            If (chars Is Nothing) Then
                Throw New ArgumentNullException("chars")
            End If

            If (chars.Length = 0) Then
                Return 0
            End If

            If (byteIndex < 0) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, String.Format(My.Resources.CannotBeNegative, "byteIndex"))
            End If

            If (charIndex < 0) Then
                Throw New ArgumentOutOfRangeException("charIndex", charIndex, String.Format(My.Resources.CannotBeNegative, "charIndex"))
            End If

            If (charCount < 0) Then
                Throw New ArgumentOutOfRangeException("charCount", charCount, String.Format(My.Resources.CannotBeNegative, "charCount"))
            End If

            If (chars.Length - charIndex < charCount) Then
                Throw New ArgumentOutOfRangeException("", charIndex + charCount, String.Format(My.Resources.IndexAndCountAreNotValidInArray, "charIndex", "charCount", "chars"))
            End If

            If (bytes.Length - byteIndex < charCount) Then
                Throw New ArgumentException(String.Format(My.Resources.NotEnouchCapacity, "bytes", "byteIndex", "bytes"))
            End If

            If (byteIndex >= bytes.Length) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, String.Format(My.Resources.IndexNotValidInArray, "byteIndex", "bytes"))
            End If

            If (charIndex >= chars.Length) Then
                Throw New ArgumentOutOfRangeException("charIndex", byteIndex, String.Format(My.Resources.IndexNotValidInArray, "charIndex", "chars"))
            End If

            '*
            '* Starts several threads in order to speed up the conversion.
            '*
            __ThreadsGetBytes = 0
            For i As Integer = 0 To charCount Step 256
                Dim state As New stateObjectStruct
                With state
                    .byteIndex = byteIndex + i
                    .charIndex = charIndex + i
                    .charArray = chars
                    .byteArray = bytes

                    .Count = Math.Min(256, (charCount - i))
                End With

                __ThreadsGetBytes += 1
                Do While Not System.Threading.ThreadPool.QueueUserWorkItem(AddressOf Me.GetBytesUtility, state)
                    '*
                    '* Release the application to process events
                    '*
                    System.Threading.Thread.Sleep(10)
                Loop
            Next

            '*
            '* Wait until all threads are finished
            '*
            Do While (__ThreadsGetBytes > 0)
                '*
                '* Release the application to process events
                '*
                System.Threading.Thread.Sleep(10)
            Loop

            Return (charCount)

        End Function

        ''' <summary>Calculates the number of characters produced by decoding a sequence of bytes from the specified byte array.</summary>
        ''' <returns>The number of characters produced by decoding the specified sequence of bytes.</returns>
        ''' <param name="count">The number of bytes to decode. </param>
        ''' <param name="index">The index of the first byte to decode. </param>
        ''' <param name="bytes">The byte array containing the sequence of bytes to decode. </param>
        ''' <exception cref="T:System.ArgumentNullException">bytes is null.</exception>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">index or count is less than zero.
        ''' <para>- or -</para>index and count do not denote a valid range in bytes.</exception>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetCharCount(ByVal bytes() As Byte, ByVal index As Integer, ByVal count As Integer) As Integer
            '*
            '* Check all the parameters
            '*
            If (bytes Is Nothing) Then
                Throw New ArgumentNullException("bytes")
            End If

            If (index < 0) Then
                Throw New ArgumentOutOfRangeException("index", index, String.Format(My.Resources.CannotBeNegative, "index"))
            End If

            If (index + count > bytes.Length) Then
                Throw New ArgumentOutOfRangeException("", index + count, String.Format(My.Resources.IndexAndCountAreNotValidInArray, "index", "count", "bytes"))
            End If

            Return (count)
        End Function

        ''' <summary>Decodes a sequence of bytes from the specified byte array into the specified character array.</summary>
        ''' <returns>The actual number of characters written into chars.</returns>
        ''' <param name="chars">The character array to contain the resulting set of characters. </param>
        ''' <param name="byteCount">The number of bytes to decode. </param>
        ''' <param name="byteIndex">The index of the first byte to decode. </param>
        ''' <param name="charIndex">The index at which to start writing the resulting set of characters. </param>
        ''' <param name="bytes">The byte array containing the sequence of bytes to decode. </param>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">byteIndex or byteCount or charIndex is less than zero.
        ''' <para>- or -</para>byteindex and byteCount do not denote a valid range in bytes.
        ''' <para>- or -</para>charIndex is not a valid index in chars. </exception>
        ''' <exception cref="T:System.ArgumentNullException">bytes is null.
        ''' <para>- or -</para>chars is null.</exception>
        ''' <exception cref="T:System.ArgumentException">chars does not have enough capacity from charIndex to the end of the array to accommodate the resulting characters.</exception>
        ''' <remarks>
        ''' Each byte in the array is converted to a character through the <see cref="M:Microsoft.VisualBasic.Chr"/> function.
        ''' </remarks>
        <DebuggerStepThrough()> _
        Public Overloads Overrides Function GetChars(ByVal bytes() As Byte, ByVal byteIndex As Integer, ByVal byteCount As Integer, ByVal chars() As Char, ByVal charIndex As Integer) As Integer
            '*
            '* Check all the parameters
            '*
            If (bytes Is Nothing) Then
                Throw New ArgumentNullException("bytes")
            End If

            If (chars Is Nothing) Then
                Throw New ArgumentNullException("chars")
            End If

            If (bytes.Length = 0) Then
                Return 0
            End If

            If (byteIndex < 0) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, String.Format(My.Resources.CannotBeNegative, "byteIndex"))
            End If

            If (byteCount < 0) Then
                Throw New ArgumentOutOfRangeException("byteCount", byteCount, String.Format(My.Resources.CannotBeNegative, "byteCount"))
            End If

            If (charIndex < 0) Then
                Throw New ArgumentOutOfRangeException("charIndex", charIndex, String.Format(My.Resources.CannotBeNegative, "charIndex"))
            End If

            If (bytes.Length - byteIndex < byteCount) Then
                Throw New ArgumentOutOfRangeException("", byteIndex + byteCount, String.Format(My.Resources.IndexAndCountAreNotValidInArray, "byteindex", "byteCount", "bytes"))
            End If

            If (chars.Length - charIndex < byteCount) Then
                Throw New ArgumentException(String.Format(My.Resources.NotEnouchCapacity, "chars", "charIndex", "chars"))
            End If

            If (byteIndex >= bytes.Length) Then
                Throw New ArgumentOutOfRangeException("byteIndex", byteIndex, String.Format(My.Resources.IndexNotValidInArray, "byteIndex", "bytes"))
            End If

            If (charIndex >= chars.Length) Then
                Throw New ArgumentOutOfRangeException("charIndex", byteIndex, String.Format(My.Resources.IndexNotValidInArray, "charIndex", "chars"))
            End If

            '*
            '* Starts several threads in order to speed up the conversion.
            '*
            __ThreadsGetChars = 0
            For i As Integer = 0 To bytes.Length Step 256
                Dim state As New stateObjectStruct
                With state
                    .byteIndex = byteIndex + i
                    .charIndex = charIndex + i
                    .charArray = chars
                    .byteArray = bytes
                    .Count = Math.Min(256, (byteCount - i))
                End With

                __ThreadsGetChars += 1
                Do While Not System.Threading.ThreadPool.QueueUserWorkItem(AddressOf Me.GetCharsUtility, state)
                    '*
                    '* Release the application to process events
                    '*
                    System.Threading.Thread.Sleep(10)
                Loop
            Next

            '*
            '* Wait until all threads are finished
            '*
            Do While (__ThreadsGetChars > 0)
                '*
                '* Release the application to process events
                '*
                System.Threading.Thread.Sleep(10)
            Loop

            Return (byteCount)

        End Function

        ''' <summary>Calculates the maximum number of bytes produced by encoding the specified number of characters.</summary>
        ''' <returns>The maximum number of bytes produced by encoding the specified number of characters.</returns>
        ''' <param name="charCount">The number of characters to encode. </param>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">charCount is less than zero. </exception>
        <DebuggerStepThrough()> _
        Public Overrides Function GetMaxByteCount(ByVal charCount As Integer) As Integer
            Return charCount
        End Function

        ''' <summary>When overridden in a derived class, calculates the maximum number of characters produced by decoding the specified number of bytes.</summary>
        ''' <returns>The maximum number of characters produced by decoding the specified number of bytes.</returns>
        ''' <param name="byteCount">The number of bytes to decode. </param>
        ''' <exception cref="T:System.ArgumentOutOfRangeException">byteCount is less than zero. </exception>
        <DebuggerStepThrough()> _
        Public Overrides Function GetMaxCharCount(ByVal byteCount As Integer) As Integer
            Return byteCount
        End Function

        ''' <summary>Gets a value indicating whether the current encoding uses single-byte code points.</summary>
        ''' <returns>true if the current <see cref="T:System.Text.Encoding"></see> uses single-byte code points; otherwise, false.</returns>
        Public Overrides ReadOnly Property IsSingleByte() As Boolean
            <DebuggerStepThrough()> _
            Get
                Return True
            End Get
        End Property

        ''' <summary>
        ''' Utility function used by GetBytes to speed up the encoding.
        ''' </summary>
        ''' <param name="stateObject">An object that contains the data to process.</param>
        <DebuggerStepThrough()> _
        Private Sub GetBytesUtility(ByVal stateObject As Object)
            Dim x As stateObjectStruct = CType(stateObject, stateObjectStruct)

            For i As Integer = 0 To (x.Count - 1)
                x.byteArray(x.byteIndex + i) = Asc(x.charArray(x.charIndex + i))
            Next

            __ThreadsGetBytes -= 1
        End Sub

        ''' <summary>
        ''' Utility function used by GetChars to speed up the encoding.
        ''' </summary>
        ''' <param name="stateObject">An object that contains the data to process.</param>
        <DebuggerStepThrough()> _
        Private Sub GetCharsUtility(ByVal stateObject As Object)
            Dim x As stateObjectStruct = CType(stateObject, stateObjectStruct)

            For i As Integer = 0 To (x.Count - 1)
                x.charArray(x.charIndex + i) = Chr(x.byteArray(x.byteIndex + i))
            Next

            __ThreadsGetChars -= 1
        End Sub

    End Class

End Namespace
